'''
This script finds the mean chat_rate per unique user per `min_` min chunk,
isolates to each `min_` timestamp, and sorts the resulting df by largest number
'''
import pandas as pd
import datetime as dt
from .helpers import data_handler as d


def thalamus(dataframe, min_):
    '''
    Formats data for rate_finder(), gets chunk_list to pass through rate_finder
    '''
    # split into hours
    dfs = d.dfSplitter(dataframe)
    dfs.find_rest()
    hour_list = dfs.result
    first_stamp = hour_list[0]
    del hour_list[0]
    # label each dataset so we can tell what hour they came from
    for i in range(len(hour_list)):
        hour_list[i]['hour'] = i

    # split hours into 2 minute chunks
    chunk_list = []
    for hour in hour_list:
        xmc = d.xminChats(hour, dataframe['_id'].unique(), min_=2) # split into 2 min segments
        xmc.find_rest()

        for chunk in xmc.result:
            chunk_list.append(chunk) # fmc.result gets returned as a list
            
    chat_rates = pd.DataFrame(columns = ['hour', 'chunk', 'start', 'end', '_id', 'num_chats', 'chats_per_2min'])

    # for each 2 min chunk
    for chunk in chunk_list:
        hour = chunk.iloc[-1,12] # col 12 is hour.
        # find the chat rate for each user
        chat_rates = chat_rates.append(rate_finder(dataframe = chunk, hour = hour, x=2)) 

    chat_rates = chat_rates.reset_index(drop=True)
    chat_rates.sort_values('chats_per_2min', ascending=False).head()
    
    return chat_rates, first_stamp
    
    
def rate_finder(dataframe, hour, x=2):
    '''
    Finds the rate of messages sent per X minutes for each user in the dataframe.
    
    NOTE: if only 1 timestamp in chunk dataframe, assumes the chunk is exactly 2 minutes before the next chunk in the entire twitch chat stream
    '''
    # Initiate new df
    chat_rate_df = pd.DataFrame(columns = ['hour','chunk','start','end','_id','num_chats', f'chats_per_{x}min'])
    chatters = dataframe['_id'].unique() # id unique chatters for this chunk
    
    
    for i in range(len(chatters)):
        temp_df = dataframe[dataframe['_id']==chatters[i]] # isolate chatter's data
        hour = hour
        chunk = i
        time_start = dataframe.iloc[0,0]
        time_end = dataframe.iloc[-1,0]
        _id = chatters[i]
        num_chats = len(temp_df['body']) # count how many chats were sent by them
        time_d = dt.timedelta.total_seconds(dataframe.iloc[-1,0] - dataframe.iloc[0,0]) 

        if time_d > 0:
            chat_rate = (num_chats/time_d) * 60 * x # use time_d to calculate chat/sec, then multiply to get user requested rate
        elif time_d == 0:
            # if there is only 1 message in the chunk, there will be only 1 timestamp
            # in that case assume that time_d = 2
            time_d = 2
            chat_rate = (num_chats/time_d) * 60 * x # convert to chat/X minutes
        else:
            chat_rate = -100 # if number is negative, math is wrong somewhere and needs to be looked into

        # gather results
        results = pd.DataFrame({
            'hour':[hour],
            'chunk':[chunk],
            'start':[time_start],
            'end':[time_end],
            '_id':[_id],
            'num_chats':[num_chats],
            f'chats_per_{x}min':[chat_rate]
        })

        chat_rate_df = chat_rate_df.append(results) # store in df
    return chat_rate_df.reset_index(drop=True)
      
    
def run(data, min_=2, save_json = False):
    '''
    Runs algo2 to find the mean chat_rate per unique user per `min_` chunk,
    takes the means for each chunk, and then sorts by the highest mean rate.
    
    input
    ------
    data: list
        List of dictionaries of data from Twitch chat
    min_: int
        Approximate number of minutes each clip should be
    save_json: bool
        True if want to save results as json to exports folder
        
    output
    ------
    json_results: list
        List of dictionaries in json format, ordered from predicted best to worst candidates.
            Ex: [{start:TIMESTAMP_INT, end:TIMESTAMP_INT}]
    '''
    data = pd.DataFrame.from_records(data)
    big_df = d.organize_twitch_chat(data) # fetch appropriate data
    results, first_stamp = thalamus(big_df, min_)
    json_results = d.results_jsonified(results,first_stamp, f'chats_per_{min_}min')
    
    if save_json:
        d.save_json(json_results, f'algo2_mean_rate_per_{min_}min')
    
    return json_results



'''
This script finds the number of words/emojis/both depending on `goal` variable,
isolates to each `min_` timestamp, and sorts the resulting df by largest number
'''
import pandas as pd
from .helpers import data_handler as d


def thalamus(big_df, min_, goal='num_words'):
    '''
    Calculates num_words/emoji/both then runs through functions to 
    split big_df, format for saving, and save as json
    
    input
    -----
    big_df: pd.DataFrame
        twitch stream chat dataframe
    min_: int
        number of minutes each chunk should be
    goal: str
        one of `num_words`, `num_emo`, or `num_words_emo`
    '''
    # cut
    big_df = algorithm(big_df)
    first_stamp, chunk_list = d.get_chunks(big_df)
    chunk_df = pd.DataFrame(columns = chunk_list[0].columns)
    for chunk in chunk_list:
        chunk_df = chunk_df.append(chunk)
        
    results = results_formatter(chunk_df, goal=goal)
    return results, first_stamp
    
def algorithm(big_df):
    # find number of words+emojis
    big_df = num_words_in_chat(big_df)
    # find number of emojis
    big_df['num_emo'] = big_df['emoticons'].apply(lambda x: len(x) if type(x) == list else 0)
    # find number of words only
    big_df['num_words'] = big_df['num_words_emo'] - big_df['num_emo']
    
    return big_df
    
def results_formatter(dataframe, goal):
    '''
    Creates a new df `results` that contains the total number of words in the dataframe, the time
    the time the dataframe started and ended
    input
    -----
    dataframe: pd.DataFrame
        Dataframe with all the hours and chunks labeled, and num_words calculated
    goal: str
        Col name that has calculated results (ex: num_words, chat_rate, etc.)
    output
    ------
    results: pd.DataFrame
        Dataframe with `hour`, `chunk`, `start`, `end`, `num_words` columns
    '''
    hour_list = []
    chunk_list = []
    time_start_list = []
    time_end_list = []
    goal_list = []
    
    dataframe = dataframe.sort_values('created_at')
    # label each chunk with unique ID
    dataframe['category'] = dataframe['hour'].astype(str) + dataframe['chunk'].astype(str)
    
    for cat in dataframe['category'].unique():
        temp_df = dataframe[dataframe['category'] == cat] 
        hour_list.append(temp_df.iloc[0, -3])  # assumes hour col is 3 from end
        chunk_list.append(temp_df.iloc[0, -2])  # assumes hunk col is 2 from end
        time_start_list.append(temp_df.iloc[0, 0])  # assume created_at col is first
        time_end_list.append(temp_df.iloc[-1, 0])  # assume created_at col is first
        goal_list.append(temp_df[goal].sum())

    results = pd.DataFrame({
        'hour':hour_list,
        'chunk':chunk_list,
        'start':time_start_list,
        'end':time_end_list,
        goal:goal_list
    })
    return results

def num_words_in_chat(dataframe):
    '''
    Creates a new col `num_words_emo` that contains the number of words in each message
    '''
    word_bag = []
    body_split = dataframe['body'].str.split(' ')  # split each string at the whitespace
    for row in body_split:
        num_words = len(row)
        word_bag.append(num_words)
    
    dataframe['num_words_emo'] = word_bag
    return dataframe

def run(data, min_=2, goal='num_words_emo', save_json = False):
    '''
    Runs algo3_5 to sort timestamps by the number of words+emojis by default.
    
    input
    ------
    data: list
        List of dictionaries of data from Twitch chat
    min_: int
        Approximate number of minutes each clip should be
    goal: str
        'num_words': sum of the number of words in each chat message
        'num_emo': sum of the number of emoticons in each chat message
        'num_words_emo': sum of the number of words + emoticons in each chat message
    save_json: bool
        True if want to save results as json to exports folder
        
    output
    ------
    json_results: list
        List of dictionaries in json format, ordered from predicted best to worst candidates.
            Ex: [{start:TIMESTAMP_INT, end:TIMESTAMP_INT}]
    '''
    data = pd.DataFrame.from_records(data)
    big_df = d.organize_twitch_chat(data) # fetch appropriate data
    results, first_stamp = thalamus(big_df, min_, goal=goal) 
    
    json_results = d.results_jsonified(results,first_stamp, goal)
    if save_json:
        d.save_json(json_results, f"algo3.5_{goal}")
    
    return json_results


'''
DESCRIPTION
    Sorts the final results by `perc_rel_unique`. Calculated as "number of chatters at timestamp"/"number of chatters in that one hour"

HOW TO
    algo1.run(data, sort_by, min_, save_json = False)
'''

import pandas as pd
from .helpers import data_handler as d

def perc_uniques(chunk_list, min_, total_uniques, big_unique):
    '''
    Finds the percent unique chatters for each dataframe in the list. Dataframes assumed to be split using xminChats.find_rest.
    '''

    perc_unique = {
            f'{min_}min_chunk':[],
            'start':[],
            'end':[],
            'num_unique':[],
            'perc_rel_unique':[],
            'perc_abs_unique':[]
    }


    for i in range(len(chunk_list)):
        # calcuate
        chunk = i
        unique = len(chunk_list[i]['_id'].unique())
        timestamp = [chunk_list[i]['created_at'].min(), chunk_list[i]['created_at'].max()]
        perc_rel = unique/total_uniques # this is the total uniques in THAT DATAFRAME, ie the hourly cut
        perc_abs = unique/big_unique # this is the total uniques in the entire twitch session
        # store
        perc_unique[f'{min_}min_chunk'].append(chunk)
        perc_unique['start'].append(timestamp[0])
        perc_unique['end'].append(timestamp[1])
        perc_unique['num_unique'].append(unique)
        perc_unique['perc_rel_unique'].append(perc_rel)
        perc_unique['perc_abs_unique'].append(perc_abs)

    df_unique = pd.DataFrame(perc_unique)
    df_unique['elapsed'] = df_unique['end'] - df_unique['start']
    return df_unique

def hour_iterator(big_df, min_=2, sort_by='rel'):
    '''
    Pushes all dfs in a list through the xminChats function, returns a dataframe of results
    
    input
    -----
    big_df: pd.DataFrame
        Df of the entire twitch session. This is the one that was split by dfSplitter class
    min_: int
        How long a timestamp range should be
    sort_by: str
        Whether to sort values by `abs` or `rel` unique chatters.
    '''
    ds = d.dfSplitter(big_df) # initiate
    ds.find_rest() # split big_df into 1 hour long separate dfs
    hour_list = ds.result # result stored in class var. NOTE: index 0 is always the very first timestamp of big_df
    first_sec = hour_list[0]
    hour_list = hour_list[1:]

    # initiate empty results df
    results = pd.DataFrame(columns=['hour', f'{min_}min_chunk', 'start', 'end', 'num_unique', 'perc_rel_unique', 'perc_abs_unique'])
    max_uniques = len(big_df['_id'].unique()) # the total number of unique chatters for the entire twitch session

    # iterate all sections through the class
    for i in range(len(hour_list)):
        fm = d.xminChats(hour_list[i], max_uniques, min_=min_)
        _n = fm.find_rest() # _n not needed
        chunk_list = fm.result # get back list of dfs, each 2 minutes long
        
        hr_uniques = perc_uniques(chunk_list, min_, total_uniques=fm.total_uniques, big_unique=fm.big_unique)
        hr_uniques['hour'] = i + 1
        results = results.append(hr_uniques)

    results['elapsed'] = results['end'] - results['start'] # to double check length
    pretty_results = results.reset_index(drop=True) # prettify
    pretty_results = pretty_results.sort_values(f'perc_{sort_by}_unique',ascending=False)
    
    json_results = d.results_jsonified(results, first_sec, results_col=f'perc_{sort_by}_unique') # ordered by top perc_rel_unique
    
    return pretty_results, json_results


def run(data, min_=2, sort_by='rel', save_json = False):
    '''
    Runs algo1 to sort timestamps by the relative percentage of chatters by default.
    
    input:
    ------
    data: list
        List of dictionaries of data from Twitch chat
    min_: int
        Approximate number of minutes each clip should be
    sort_by: str
        'rel': "number of chatters at timestamp"/"number of chatters at that hour"
        'abs': "number of chatters at timestamp"/"total number of chatters in stream"
    save_json: bool
        True if want to save results as json to exports folder
    '''
    data = pd.DataFrame.from_records(data)
    big_df = d.organize_twitch_chat(data) # fetch appropriate data   
    results, json_results = hour_iterator(big_df,min_=min_ , sort_by=sort_by)
    if save_json:
        d.save_json(json_results, name=f"algo1_perc_{sort_by}_unique")
    return json_results



'''
    Algorithms for Pillar. Currently includes 4 "mini" algorithms with different 
    options, nothing too sophisticated.
HOW TO
    from pillaralgos import OPTIONS  # replace options with an algo
    OPTIONS.run(data)                # see algo docstring for descriptions
'''
__version__ = '1.0.6'


'''
Functions to help with graphing pretty graphs
'''

def plot_with_time(x,y,data, xformat='%H:%m'):
    '''
    Plots a scatterplot with the xaxis neatly formatted
    
    input
    -----
    x: str
        Column name. Column must be datetime
    y: str
    data: pd.DataFrame
    xformat: str
        Format xaxis should be displayed in, using strftime syntax
    '''
    import matplotlib.pyplot as plt
    import matplotlib.dates as mdates
    import seaborn as sns
    
    fig, ax = plt.subplots()
    sns.scatterplot(x=x,y=y,data=data)
    
    myFmt = mdates.DateFormatter(xformat)
    ax.xaxis.set_major_formatter(myFmt)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pillaralgos', 'pillaralgos.helpers']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.20.2,<2.0.0', 'pandas>=1.2.3,<2.0.0']

setup_kwargs = {
    'name': 'pillaralgos',
    'version': '1.0.6',
    'description': 'Algorithms for Pillar. Currently includes "mini" algorithms, nothing too sophisticated.',
    'long_description': '# Table of Contents\n1. [Build and Publish](#build)\n1. [Background](#background)\n   1. [Algorithms](#algorithms)\n   2. [Datasets](#datasets)\n3. [Current Goal](#current-goal)\n4. [Long Term Goal](#long-term-goal)\n\n# Build\nTo build and publish this package we are using the [poetry](https://python-poetry.org/) python packager. It takes care of some background stuff that led to mistakes in the past.\n\nFolder structure:\n```\n|-- pypi\n    |-- pillaralgos\n        |-- helpers\n            |-- __init__.py\n            |-- data_handler.py\n            |-- graph_helpers.py\n            |-- sanity_checks.py\n        |-- __init__.py  # must include version number\n        |-- algoXX.py\n    |-- LICENSE\n    |-- README.md\n    |-- pyproject.toml  # must include version number\n```\nTo publish just run the `poetry publish --build` command after update version numbers as needed.\n\n# Background\nPillar is creating an innovative way to automatically select and splice clips from Twitch videos for streamers. This repo is focusing on the algorithm aspect. Three main algorithms are being tested.\n\n## Algorithms\n\n1. [Algorithm 1](https://github.com/pomkos/twitch_chat_analysis/blob/reorganize_repo/algorithm_1.ipynb): Find the best moments in clips based on where the most users participated. Most is defined as the *ratio of unique users* during a 2 min section to unique users for the entire session.\n1. [Algorithm 2](https://github.com/pomkos/twitch_chat_analysis/blob/reorganize_repo/algorithm_2.ipynb) Find the best moments in clips based on when rate of messages per user peaked. This involves answering the question "at which 2 min segment do the most users send the most messages?". If users X, Y, and Z all send 60% of their messages at timestamp range delta, then that timestamp might qualify as a "best moment"\n   1. __NOTE__: Currently answers the question "at which 2 min segment do users send the most messages fastest"\n1. [Algorithm 3 (WIP)](https://github.com/pomkos/twitch_chat_analysis/blob/reorganize_repo/algorithm_3.ipynb) Weigh each user by their chat rate, account age, etc. Heavier users predicted to chat more often at "best moment" timestamps \n   1. __STATUS__: current weight determined by (`num_words_of_user`/`num_words_of_top_user`)\n   1. [Algorithm 3.5](https://github.com/pomkos/twitch_chat_analysis/blob/reorganize_repo/algorithm_3.5.ipynb) Finds the best moments in clips based on most number of words/emojis/both used in chat\n\n## Datasets:\n1. Preliminary data `prelim_df`: 545 rows representing one 3 hour 35 minute 26 second twitch stream chat of [Hearthstone by LiiHS](https://www.twitch.tv/videos/963184458)\n    * Used to create initial json import and resulting df clean/merge function `organize_twitch_chat`\n2. Big data `big_df`: 2409 rows representing one 7 hour 37 minute, 0 second twitch stream chat of [Hearthstone by LiiHS](https://www.twitch.tv/videos/955629991)\n    * Used to create all algorithms\n\n# Current Goal\n\nTo create one overarching algorithm that will find the most "interesting" clips in a twitch VOD. This will be created through the following steps:\n1. Creation of various algorithms that isolate `min_` (2 by default) minute chunks. The basic workflow:\n   1. Create variable (ex: `num_words`, for number of words in the body of a chat message)\n   1. Group df by `min_` chunks, then average/sum/etc `num_words` for each `min_` chunks\n   1. Sort new df by `num_words`, from highest "value" to lowest "value"\n   1. Return this new df as json ([example](https://github.com/pomkos/twitch_chat_analysis/blob/main/exports/algo1_results.json))\n1. Users rate clips provided by each algorithm\n2. Useless algorithms thrown away\n3. Rest of the algorithms merged into one overarching algorithm, with weights distributed based on user ratings\n\n# Long Term Goal\n\n* __New objective measure__: community created clips (`ccc`) for a given VOD id with start/end timestamps for each clip\n* __Assumption__: `ccc` are interesting and can be used to create a narrative for each VOD. We can test this by cross referencing with posts to /r/livestreamfails upvotes/comments\n* __Hypothesis__: if we can predict where `ccc` would be created, those are potentially good clips to show the user\n   * *Short term test*: Create a model to predict where ccc would be created using variables such as word count, chat rate, emoji usage, chat semantic analysis. We can do this by finding timestamps of ccc and correlating them with chat stats\n   * *Medium term test*: Use top 100 streamers as training data. What similarities do their ccc and reddit most upvoted of that VOD share? (chat rate etc)\n      1. Get the transcript for these top 100\n      2. Get the top 100\'s YT posted 15-30min story content for the 8 hour VOD\n      3. Get the transcript for that story content\n      4. Semantic analysis and correlations, etc.\n   * *Long term test*: what percentage of clips do our streamers actually end up using\n',
    'author': 'Peter Gates',
    'author_email': 'pgate89@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# this file contains a series of classes and functions to help with loading and splitting twitch chat data
# also json_saver() that converts given variable into string, saves into .json file

import pandas as pd
import numpy as np
import datetime as dt

def rename_columns(col_string):
    '''
    Renames columns to be more presentable
    '''
    if (col_string == 'created_at_id') | (col_string == 'updated_at_id'):
        col_string = col_string.replace('_id','')
        return 'id_' + col_string
    elif (col_string == 'created_at_mess') | (col_string == 'updated_at_mess'):
        col_string = col_string.replace('_mess','')
        return col_string
    else:
        return col_string.replace('_mess','',1).replace('_id','',1)

def select_columns(dataframe, keep_user_vars=False):
    '''
    Removes unneeded columns
    '''
    if keep_user_vars:
        # If true, include user columns
        bad_cols = [
            'display_name_id',
            'name_id',
            'user_notice_params_mess'
        ]
    else:
        # If false, not analyzing the users
        bad_cols = [
            'display_name_id',
            'name_id',
            'user_notice_params_mess',
        # these aren't used by the simpler fctns
            'bio_id',
            'created_at_id',
            'updated_at_id',
            'logo_id'
        ]
    dataframe = dataframe.drop(bad_cols, axis=1)
    cols = dataframe.columns
    cols = list(pd.Series(cols).apply(rename_columns))
    dataframe.columns = cols
    return dataframe

def organize_twitch_chat(data, keep_user_vars=False):
    '''
    Turns json into dataframe. Expands lists of lists into own columns.
    
    input
    -----
    data: list
        list of dictionaries in json format, loaded with the `open` context manager.
        
    output
    ------
    df: pd.DataFrame
        Dataframe with the following columns: 
            ['created_at', 'updated_at', 'display_name', '_id', 'name', 'type',
             'bio', 'logo', 'body', 'is_action', 'user_badges', 'emoticons']
    '''
    data = pd.DataFrame.from_records(data) # convert to df
    df = data[['created_at','updated_at','commenter','message']].add_suffix('_mess')
    
    h = dictExtractor(df['message_mess'],label='_mess')
    messages = h.result
    g = dictExtractor(df['commenter_mess'], label='_id')
    users = g.result
    
    df = df.drop(['message_mess','commenter_mess'], axis=1) # duplicate info
    df = pd.concat([df,users,messages],axis=1)
    # all vars were loaded as str. Change type to datetime/int/bool
    df = df.astype( {'_id_id':int,
                    'bio_id':'category',
                    'created_at_id':'datetime64[ns]',
                    'created_at_mess':'datetime64[ns]',
                    'updated_at_id':'datetime64[ns]',
                    'updated_at_mess':'datetime64[ns]',
                    'is_action_mess':bool,
                    'type_id':'category'}
                  )
    df = select_columns(df, keep_user_vars)
    return df

class dictExtractor:
    def __init__(self, my_series, label = ''):
        '''
        Extracts dictionaries from series into a new dict using the
        longest dictionary's keys. Converts new dict into df, stored
        as `self.result`. Because not every dict had same number of keys.
        
        input
        -----
        my_series: pd.Series
            A column from twitch dataframe where each row is a dict
        label: str
            What will be appended to the end of each col
        '''
        # find max length of dicts
        length = my_series.apply(lambda x: len(x))
        y = 0
        for x in length:
            if x > y:
                y = x
        # find index of max keys dict
        ind = length[length == 8].index[0]
        max_d = my_series.iloc[ind].keys()
        self.max_d = max_d
        # initiate new dict
        self.new_dict = {}
        for k in max_d:
            self.new_dict[k] = []
        # extract dict values into new dict
        my_series.apply(lambda x: self.keys_iterator(x))
        # store as df
        self.result = pd.DataFrame.from_dict(self.new_dict)
        # df.add_suffix() is actually 0.25 seconds slower
        self.result.columns = [col + label for col in self.result.columns]
        
    def keys_iterator(self, my_dict):
        '''
        Checks that all of the `max_d` are in the given dictionary. If not,
        appends np.nan. Otherwise appends the value.
        '''
        for k in self.max_d:
            if k not in my_dict.keys():
                self.new_dict[k].append(np.nan)
            else:
                self.new_dict[k].append(my_dict[k])


class dfSplitter():
    def __init__(self, dataframe):
        '''
        Splits dataframe into multiple dataframes, each 1 hour long

        output:
        ------
        my_list: list
            List of dataframes
        '''
        # init function finds the first split
        dataframe = dataframe.sort_values("created_at")
        first = dataframe[dataframe['created_at'] <= dataframe.loc[0,'created_at'] + pd.Timedelta(hours = 1)]
        self.last_i = first.index.max()
        self.dataframe = dataframe
        self.result = [] # list to append starting timestamp + datasets to
        self.result.append(dataframe.iloc[0, 0]) # NOTE: assumes first col is always "created_at" col
        self.result.append(first)
        
    def find_rest(self):
        '''
        Uses last index of first split to find the others
        '''
        dataframe = self.dataframe
        last_i = self.last_i
        if last_i+1 != len(dataframe):
            new_df = dataframe.loc[last_i+1:,:] # clip df to start at last_i
            newest = new_df[new_df['created_at'] <= new_df.loc[last_i+1,'created_at'] + pd.Timedelta(hours=1)] # filter by hour
            self.result.append(newest) # store in list
            self.last_i = newest.index.max()
            
            self.find_rest() # repeat
        else:
            return dataframe # never actually used
        
class xminChats():
    def __init__(self,dataframe, big_unique, min_= 2):
        '''
        Finds the percent unique chatters that chatted every min_ minutes
        
        input
        -----
        dataframe: pd.DataFrame
            Twitch chat dataframe organized and split by dfSplitter
        big_unique: int
            Total number of unique chatters for the entire Twitch stream
        min_: int
            Minute range to find timestamps for. Ex: Find 2 min long timestamps.
        '''
        
        # init function finds the first split
        dataframe = dataframe.sort_values("created_at")
        first = dataframe[dataframe['created_at'] <= dataframe.iloc[0,0] + pd.Timedelta(minutes = min_)]
        
        self.min_ = min_
        self.total_uniques = len(dataframe['_id'].unique())
        self.big_unique = big_unique
        
        self.last_i = first.index.max()
        self.dataframe = dataframe
        
        self.result = []
        self.result.append(first)
        
    def find_rest(self):
        '''
        Uses last index of first split to find the others
        '''
        dataframe = self.dataframe
        last_i = self.last_i
        if last_i+1 < dataframe.index.max(): # NOT len(dataframe), that bugs out and i dont wanna explain why
            new_df = dataframe.loc[last_i+1:,:] # clip df to start new min_ min calc at last_i+1
            newest = new_df[new_df['created_at'] <= new_df.loc[last_i+1,'created_at'] + pd.Timedelta(value=self.min_, unit='minutes')] # filter by minute
            self.result.append(newest) # store in list
            
            self.last_i = newest.index.max()
            self.find_rest() # repeat
        else:
            x=''
            
def get_chunks(dataframe, min_=2):
    '''
    Iterates through the data_helper classes to divide dataframe into chunks
    
    input
    -----
    dataframe: pd.DataFrame
        The entire twitch stream chat df
    min_: int
        The min_ value to pass into xminChats()
        
    output
    ------
    first_stamp: datetime
        The very first timestamp of dataframe
    chunk_list:
        List of `min_` long dataframes
    '''
    dhs = dfSplitter(dataframe)
    dhs.find_rest()
    hour_list = dhs.result

    first_stamp = hour_list[0]
    del hour_list[0]

    chunk_list = []
    for i in range(len(hour_list)):
        hour = hour_list[i]
        
        dhx = xminChats(hour, dataframe['_id'].unique(), min_=min_)
        dhx.find_rest()
        chunks = dhx.result

        for x in range(len(chunks)):
            chunk = chunks[x]
            chunk['hour'] = i
            chunk['chunk'] = x
            chunk_list.append(chunk)
    return first_stamp, chunk_list

def results_jsonified(results, first_sec, results_col):
    '''
    Converts timestamps to seconds, extracts results and makes the whole thing machine readable
    
    input
    -----
    results: pd.DataFrame
        DataFrame with at least the start (datetime) and end (datetime) columns, and a column to sort by.
    first_sec: datetime
        The very first timestamp in the entire twitch chat log. Used to calculate elapsed time in seconds.        
    results_col: str
        Column to sort values by (ascending=False)
        
    output
    ------
    json_results: list
        List of dictionaries with startTime and endTime keys, sorted by best results at top
    '''
    results['first_sec'] = first_sec # to calculate elapsed time from first sec, in seconds
    results = results.sort_values(results_col, ascending=False) # so json format is returned with top result being the most relevant
    json_results = []
    for i, row in results.iterrows():
        og = row['first_sec']
        start = row['start']
        end = row['end']
        
        start_sec = dt.timedelta.total_seconds(start-og) # find difference between first sec and given timestamp, convert that to seconds
        end_sec = dt.timedelta.total_seconds(end-og)
        
        dict_ = {"startTime":start_sec,
                 "endTime":end_sec}
        json_results.append(dict_)
        
    return json_results

def save_json(json_results, name):
    '''
    Converts list of dict to pure str, then saves as a json file.
    
    input
    -----
    json_results: list
        List of dictionaries containing results
    name: str
        Filename (with optional directory) to save as. Ex: name.json or exports/name.json
    '''
    str_  = '['
    for dict_ in json_results:
        str_ += str(dict_) + ', \n '
    str_ += ']'
    
    with open(f"{name}.json",'w') as f:
        f.write(str_)
    print(f"Saved to {name}.json")


'''
This script finds the top 10 active users, timestamps where they participated,
filtered by at least `min_words` number of words sent by the user per stamp
'''
import pandas as pd
from .helpers import data_handler as dh

def thalamus(big_df, min_, goal, min_words):
    
    id_words = id_words_counter(big_df)
    first_stamp, chunk_list = dh.get_chunks(big_df)
    top_chunks = new_chunk_list(id_words, chunk_list, min_words = 5)
    results = results_formatter(top_chunks, goal='num_top_user_appears')
    
    return results, first_stamp
    
def results_formatter(list_chunk, goal):
    '''
    Creates a new df `results` that contains the total number of words in the dataframe, the time
    the time the dataframe started and ended
    input
    -----
    list_chunk: list
        List of pd.DataFrame
    goal: str
        Col name that has calculated results (ex: num_words, chat_rate, etc.)
    output
    ------
    results: pd.DataFrame
        Dataframe with `start`, `end`, `num_words` columns
    '''
    chunk_list = []
    time_start_list = []
    time_end_list = []
    goal_list = []
    
    for chunk in list_chunk:
        time_start_list.append(chunk.iloc[0, 0])  # assume created_at col is first
        time_end_list.append(chunk.iloc[-1, 0])  # assume created_at col is first
        goal_list.append(chunk[goal].sum())

    results = pd.DataFrame({
        'start':time_start_list,
        'end':time_end_list,
        goal:goal_list
    })
    results = results.sort_values(goal, ascending=False).drop_duplicates()
    return results
    
def new_chunk_list(id_words, chunk_list, min_words):
    '''
    Creates a new list of chunks, containing only chunks where top
    users sent more than `min_words` words
    '''
    user_dict = {} # store chunks of the top 10 users
    for user in id_words.head(10)['_id']:
        user_dict[user] = []
    for user in user_dict.keys():
        for chunk in chunk_list:
            chunk['_id'] = chunk['_id'].astype(int)
            if sum(chunk['_id'] == user) > min_words:
                user_dict[user].append(chunk)
    top_chunks = []
    for user, value in user_dict.items():
        if len(value) > 0:
            for chunk in value:
                chunk['num_top_user_appears'] = chunk['_id']==user
                top_chunks.append(chunk)
    return top_chunks

def id_words_counter(big_df):
    '''
    Returns a dataframe with all user IDs and the number of words/emojis/combined
    they each sent, sorted by top senders
    '''
    id_words = pd.DataFrame(columns = ['display_name','_id','num_words', 'num_emoji'])
    
    for _id in big_df['_id'].unique():
        temp_df = big_df[big_df['_id'] == _id]
        words = temp_df['body'].str.split(' ')
        emoji = temp_df['emoticons'].apply(lambda x: 0 if type(x) == float else len(x))
        num_emoji = emoji.sum()
        num_words = words.apply(lambda x: len(x))

        sum_words = num_words.sum()
        id_words = id_words.append({
            'display_name':temp_df['display_name'].iloc[0],
            '_id':_id,
            'num_words':sum_words,
            'num_emoji':num_emoji
        }, ignore_index=True)
    id_words = id_words.astype({'_id':int,'num_words':int,'num_emoji':int})
    id_words['only_words'] = id_words['num_words'] - id_words['num_emoji']
    id_words = id_words.sort_values('only_words',ascending=False).reset_index(drop=True).reset_index()
    id_words = id_words.rename({'index':'rank'},axis=1)
    # Weight = "num words / num words of biggest spammer"
    id_words['weight'] = id_words['only_words'] / id_words['only_words'].max()
    
    return id_words
    
def run(data, min_=2, min_words=5, save_json = False):
    '''
    Runs algo3_0 to extract only those chunks where the top 10 users participated. 
      - Top users are defined as "sent the most words in the entire twitch stream".
      - Once top users are identified, only those chunks are returned where top users
        sent at least `min_words` number of words.
    
    input
    ------
    data: list
        List of dictionaries of data from Twitch chat
    min_: int
        Approximate number of minutes each clip should be
    min_words:int
        When filtering chunks to top users, at least how many words the top user should send
    save_json: bool
        True if want to save results as json to exports folder
        
    output
    ------
    json_results: list
        List of dictionaries in json format, ordered from predicted best to worst candidates.
            Ex: [{start:TIMESTAMP_INT, end:TIMESTAMP_INT}]
    '''
    data = pd.DataFrame.from_records(data)
    big_df = dh.organize_twitch_chat(data) # fetch appropriate data
    results, first_stamp = thalamus(big_df, min_, min_words = 5, goal='num_top_user_appears') 
    
    json_results = dh.results_jsonified(results, first_stamp, results_col='num_top_user_appears')
    if save_json:
        dh.save_json(json_results, f"algo3.0_top_user_appears")
    
    return json_results
